/*
 * file: event.c
 * This file is part of the libpv library examples.
 *
 * Copyright 2005 Pengutronix
 * Juergen Beisert j.beisert@pengutronix.de
 *
 * Please send bug reports to bugs@pengutronix.de
 *
 * See COPYING for the status of this software.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdint.h>
#include <string.h>
#include <signal.h>

#include <libpv.h>

static int		loop=1;
static uint8_t	*random_var;
static int		device_node;

/**
 * Thread to read data from a device. It blocks until nothing could be
 * read in. When a new value arrievs save it and notify about this new value.
 **/
static void *device_thread(void *p)
{
	pv_engine_t *pv_engine=(pv_engine_t*)p;
	uint8_t var;

	while (loop) {
		if (read(device_node,&var,sizeof(var))==1) {
			pv_lock(pv_engine);
			*random_var=var;		/* lock - modify - unlock */
			pv_unlock(pv_engine);
			pv_event_generate(pv_engine,random_var);	/* notify about new value */
		}
	}
	return(NULL);
}

/**
 * to stop the endless loops
 **/
void signalhandler (int sig)
{
	if (sig == SIGTERM || sig == SIGINT)
		loop=0;	/* end the loop in main() */
}

int main(int argc,char *argv[])
{
	pv_engine_attr_t	attr;
	pv_engine_t			pv_engine;
	void				*var;
	pthread_t			thread=0;

	if (argc > 1) {
		/*
		* parse the given parameters
		*/
		if (pv_engine_attr_init_from_cmdline(&attr, argc, argv))
			return(EXIT_FAILURE);
	}
	else {
		/*
		 * Without any parameters setup default values
		 */
		pv_engine_attr_init(&attr);
		strcpy(attr.desc_filename,argv[0]);
		strcat(attr.desc_filename,".xml");
		strcpy(attr.store_filename,argv[0]);
		strcat(attr.desc_filename,".bin");
		attr.jvisu_port=0;
	}

	/*
	 * establish connection to the process image
	 */
	if (pv_init(&pv_engine, &attr))
		return(EXIT_FAILURE);

	/*
	 * Register our sole variable this simple model supports
	 */
	if ((random_var=pv_register_uint8(&pv_engine,"ucData"))) {
		/*
		 * request a change notify to this variable
		 */
		if (!pv_event_register(&pv_engine,random_var)) {

			device_node=open("/dev/random",O_RDONLY);
			if (device_node > 0) {
				signal(SIGTERM, signalhandler);
				signal(SIGINT, signalhandler);
				/*
				* create a thread to read from IO blocking device
				*/
				if (!pthread_create(&thread,NULL,device_thread,&pv_engine)) {
					while (loop) {
						pv_event_await_syncronous(&pv_engine,&var);
						if (var == random_var) {
							pv_lock(&pv_engine);
							printf("Process variable ucData was updated to 0x%02X\n",*random_var);
							pv_unlock(&pv_engine);
						}
					};
				}
			}
		}
	}
	if (pthread_cancel(thread) == 0)	/* only to be sure */
		pthread_join(thread,NULL);
	close(device_node);
	pv_destroy(&pv_engine);
	return(EXIT_SUCCESS);
}

/*
#
# vim settings
#
# vim:set ts=4:
# vim:set shiftwidth=4:
# vim:set nowrap:
#
*/

/* end if file simple.c */
